import wx
import sys
import os
import math
import processing

from NotebookEX import *
from HighlighterPanel import *
from HighlighterMonitor import *
from wxConsole import *

#TODO: Maybe always have the run button enabled, and just say "Hey, if you run, you'll kill what was running before." when run is clicked again?

class Highlighter(wx.Frame):
    """An extension of wxFrame, this class creates a window containing a wxNotebook.
The interface created by this class allows for the viewing of in-flight Python code,
with active lines highlighted in real time."""

    SLIDER_MAX = 100
    
    OPEN_TEXT_SHORT = "Open"
    OPEN_TEXT_LONG = "Open a file for highlighting"
    CLOSE_TAB_TEXT_SHORT = "Close Tab"
    CLOSE_TAB_TEXT_LONG = "Close the current tab"
    RUN_TEXT_SHORT = "Run Tab"
    RUN_TEXT_LONG = "Run the file shown in the current tab"
    STEP_TEXT_SHORT = "Step Forward"
    STEP_TEXT_LONG = "Unpause the program until a line in an open file is encountered"
    ABORT_TEXT_SHORT = "Abort Execution"
    ABORT_TEXT_LONG = "Abort the current running program"
    EXIT_TEXT_SHORT = "Exit"
    EXIT_TEXT_LONG = "Exit PyLighter and abort all running code"
    
    VERSION_NUMBER = 0.4
    DESCRIPTION_TEXT = "PyLighter Python Code Highlighting Utility\nWritten in 2008 by Michael Boland\n\nOriginal idea by Dr. Curt Clifton\n\nWritten at Rose-Hulman Institute of Technology"
    DEVELOPERS_DEVELOPERS_DEVELOPERS_DEVELOPERS = ["Michael Boland"]
    LICENSE_TEXT = """Copyright (c) 2008 Michael G. Boland, Curtis Clifton, Rose-Hulman Institute of Technology

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""

    def __init__(self, parent=None, id=-1, title="PyLighter"):
        #Call the parent constructor
        wx.Frame.__init__(self, parent, id, title, size=(600,800))
        
        #Set the minimum size
        self.SetMinSize((300,500))

        #Create the status bar at the bottom and set its initial text
        self.tb_stebatus_bar = self.CreateStatusBar()
        self.tb_stebatus_bar.SetStatusText("PyLighter")

        #Generate the menu bar and tool bar
        self.generate_menu_bar()
        self.generate_tool_bar()
        
        #Disable stuff that can't be done initially
        self.disable_run()
        self.disable_step()
        self.disable_abort()
        self.disable_close_tab()
        
        #Create the main panel
        self.main_panel = wx.Panel(self,-1)
        
        #Create the splitter and its sizer
        self.splitter = wx.SplitterWindow(self.main_panel, -1, style = wx.SP_LIVE_UPDATE)
        self.splitter.SetMinimumPaneSize(150)
        
        main_sizer = wx.BoxSizer()
        main_sizer.Add(self.splitter,1,wx.EXPAND)
        self.main_panel.SetSizer(main_sizer)

        #Create the notebook and the panel to contain it, put them in the splitter
        self.nb_panel = wx.Panel(self.splitter, style=wx.SUNKEN_BORDER)
        self.notebook = NotebookEX(self.nb_panel)

        #Create a sizer to make the notebook fill the panel
        nb_sizer = wx.BoxSizer()
        nb_sizer.Add(self.notebook,1,wx.EXPAND|wx.ALL,5)
        self.nb_panel.SetSizer(nb_sizer)
        
        #Create the console and the panel to contain it, put them in the splitter
        self.console_panel = wx.Panel(self.splitter,style=wx.SUNKEN_BORDER)
        self.console = wxConsole(self.console_panel,-1)
        
        #Create a sizer to make the console fill the panel
        c_sizer = wx.BoxSizer()
        c_sizer.Add(self.console,1,wx.EXPAND)
        self.console_panel.SetSizer(c_sizer)
        
        #Split the splitter
        self.splitter.SplitHorizontally(self.nb_panel, self.console_panel, -1)

        #Create the image list, load the images into it, and give the list to the notebook
        #These are the images used for the status lights in the tabs
        img_list = wx.ImageList(24,24)
        tab_images = []
        filenames = os.listdir("./tabimages")
        filenames.sort()
        for filename in filenames:
            if ".png" in filename:
                b = wx.Bitmap("./tabimages/%s"%filename)
                tab_images.append(b)
        for image in tab_images:
            the_index = img_list.Add(image)
        self.img_count = len(tab_images)
        self.notebook.AssignImageList(img_list)

        #Bind the close function
        self.Bind(wx.EVT_CLOSE, self.OnClose)
        
        #Set up drag-and-drop
        self.drop_target = TabOpenDropTarget(self)
        self.SetDropTarget(self.drop_target)

        #Store the about info
        self.about_info = wx.AboutDialogInfo()
        self.about_info.SetName("PyLighter")
        self.about_info.SetVersion(str(Highlighter.VERSION_NUMBER))
        self.about_info.SetDescription(Highlighter.DESCRIPTION_TEXT)
        self.about_info.SetLicense(Highlighter.LICENSE_TEXT)
        for developer in Highlighter.DEVELOPERS_DEVELOPERS_DEVELOPERS_DEVELOPERS:
            self.about_info.AddDeveloper(developer)

        #Store the running state
        self.is_running = False
        
        #Create the main_pipe variable that will be assigned later
        self.main_pipe = None
        
        #Store the current slowing duration
        self.slowing_duration = 0
        
        #Do the icon
        self.icon = wx.Icon("./pylighter-icon.ico", wx.BITMAP_TYPE_ICO)
        self.SetIcon(self.icon)

    def generate_tool_bar(self):
        """Generates the tool bar for this frame."""
        #Create the tool bar
        self.tool_bar = self.CreateToolBar()
        
        #Get the size for resizing images
        size = self.tool_bar.GetToolBitmapSize()
        if size[0] < 24 or size[1] < 24:
            size = (24,24)
            self.tool_bar.SetToolBitmapSize(size)
        
        #Load, resize, and convert the images
        open_img = wx.Image("./toolimages/open.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()
        close_img = wx.Image("./toolimages/close.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()
        run_img = wx.Image("./toolimages/run.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()
        step_img = wx.Image("./toolimages/step.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()
        abort_img = wx.Image("./toolimages/abort.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()
        exit_img = wx.Image("./toolimages/exit.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()

        op = self.tool_bar.AddLabelTool(-1,Highlighter.OPEN_TEXT_SHORT,open_img,shortHelp=Highlighter.OPEN_TEXT_SHORT,longHelp=Highlighter.RUN_TEXT_LONG)
        self.Bind(wx.EVT_TOOL, self.OnOpen, op)
        cl = self.tool_bar.AddLabelTool(-1,Highlighter.CLOSE_TAB_TEXT_SHORT,close_img,shortHelp=Highlighter.CLOSE_TAB_TEXT_SHORT,longHelp=Highlighter.CLOSE_TAB_TEXT_LONG)
        self.Bind(wx.EVT_TOOL, self.OnCloseTab, cl)
        self.tool_bar.AddSeparator()

        rn = self.tool_bar.AddLabelTool(-1, Highlighter.RUN_TEXT_SHORT, run_img,shortHelp=Highlighter.RUN_TEXT_SHORT,longHelp=Highlighter.RUN_TEXT_LONG)
        self.Bind(wx.EVT_TOOL, self.OnRunCurTab, rn)
        self.tool_bar.AddSeparator()
        
        size = self.tool_bar.GetToolBitmapSize()
        size = (size[0]*4, size[1])
        self.tb_slider = wx.Slider(self.tool_bar, -1, Highlighter.SLIDER_MAX, 0, Highlighter.SLIDER_MAX, size=size, style=wx.SL_HORIZONTAL)
        self.Bind(wx.EVT_SLIDER, self.OnTBSliderUpdate)
        self.tool_bar.AddControl(self.tb_slider)
        
        st = self.tool_bar.AddLabelTool(-1,Highlighter.STEP_TEXT_SHORT,step_img,shortHelp=Highlighter.STEP_TEXT_SHORT,longHelp=Highlighter.STEP_TEXT_LONG)
        self.Bind(wx.EVT_TOOL, self.OnStep, st)
        abort = self.tool_bar.AddLabelTool(-1,Highlighter.ABORT_TEXT_SHORT,abort_img,shortHelp=Highlighter.ABORT_TEXT_SHORT,longHelp=Highlighter.ABORT_TEXT_LONG)
        self.Bind(wx.EVT_TOOL, self.OnAbort, abort)
        self.tool_bar.AddSeparator()

        ex = self.tool_bar.AddLabelTool(-1,Highlighter.EXIT_TEXT_SHORT,exit_img,shortHelp=Highlighter.EXIT_TEXT_SHORT,longHelp=Highlighter.EXIT_TEXT_LONG)
        self.Bind(wx.EVT_TOOL, self.OnClose, ex)

        #Realize the toolbar
        self.tool_bar.Realize()

        #Store this stuff for later
        self.tb_open = op
        self.tb_close = cl
        self.tb_run = rn
        self.tb_step = st
        self.tb_abort = abort
        self.tb_exit = ex

    def OnTBSliderUpdate(self,event):
        val = Highlighter.SLIDER_MAX - float(self.tb_slider.GetValue())
        new_delay = 0
#        self.tb_slider.SetToolTipString(str(val))
        if self.is_running:
            if val == 0:
                self.disable_slowing()
                self.resume()
            elif val == Highlighter.SLIDER_MAX:
                self.pause()
            else:
                self.resume()
                self.enable_slowing()
                new_delay = math.exp(8*val/Highlighter.SLIDER_MAX)/1000
        self.set_slowing_duration(new_delay)
    
    def generate_menu_bar(self):
        """Generates the menu bar for this frame."""
        #create the menu bar
        self.menu_bar = wx.MenuBar()
        
        #Get the size for resizing images
        size = (24,24)
        
        #Load, resize, and convert the images
        open_img = wx.Image("./toolimages/open.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()
        close_img = wx.Image("./toolimages/close.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()
        run_img = wx.Image("./toolimages/run.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()
        step_img = wx.Image("./toolimages/step.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()
        abort_img = wx.Image("./toolimages/abort.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()
        exit_img = wx.Image("./toolimages/exit.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()
        about_img = wx.Image("./toolimages/about.png").Scale(size[0],size[1],wx.wx.IMAGE_QUALITY_HIGH).ConvertToBitmap()

        #create the menus
        self.file_menu = wx.Menu()
        self.run_menu = wx.Menu()
        self.help_menu = wx.Menu()

        #populate file menu
        self.menu_open = wx.MenuItem(self.file_menu, -1,"&Open\tCtrl+O",Highlighter.OPEN_TEXT_LONG)
        self.menu_open.SetBitmap(open_img)
        self.file_menu.AppendItem(self.menu_open)
        self.Bind(wx.EVT_MENU, self.OnOpen, self.menu_open)
        self.file_menu.AppendSeparator()
        self.menu_close_tab = wx.MenuItem(self.file_menu, -1,"&Close tab\tCtrl+W",Highlighter.CLOSE_TAB_TEXT_LONG)
        self.menu_close_tab.SetBitmap(close_img)
        self.file_menu.AppendItem(self.menu_close_tab)
        self.Bind(wx.EVT_MENU, self.OnCloseTab, self.menu_close_tab)
        self.file_menu.AppendSeparator()
        self.menu_exit = wx.MenuItem(self.file_menu,-1,"E&xit\tCtrl+Q",Highlighter.ABORT_TEXT_LONG)
        self.menu_exit.SetBitmap(exit_img)
        self.file_menu.AppendItem(self.menu_exit)
        self.Bind(wx.EVT_MENU, self.OnClose, self.menu_exit)

        #populate run menu
        self.menu_run = wx.MenuItem(self.run_menu,-1,"&Run current tab\tCtrl+R",Highlighter.RUN_TEXT_LONG)
        self.menu_run.SetBitmap(run_img)
        self.run_menu.AppendItem(self.menu_run)
        self.Bind(wx.EVT_MENU, self.OnRunCurTab, self.menu_run)
        self.menu_step = wx.MenuItem(self.run_menu,-1,"&Step execution\tCtrl+N",Highlighter.STEP_TEXT_LONG)
        self.menu_step.SetBitmap(step_img)
        self.run_menu.AppendItem(self.menu_step)
        self.Bind(wx.EVT_MENU, self.OnStep, self.menu_step)
        self.menu_abort = wx.MenuItem(self.run_menu,-1,"A&bort\tCtrl+B",Highlighter.ABORT_TEXT_LONG)
        self.menu_abort.SetBitmap(abort_img)
        self.run_menu.AppendItem(self.menu_abort)
        self.Bind(wx.EVT_MENU, self.OnAbort, self.menu_abort)

        #populate help menu
        self.menu_about = wx.MenuItem(self.help_menu,-1,"&About","About PyLighter")
        self.menu_about.SetBitmap(about_img)
        self.help_menu.AppendItem(self.menu_about)
        self.Bind(wx.EVT_MENU, self.OnAbout, self.menu_about)

        #add the menus to the menu bar
        self.menu_bar.Append(self.file_menu,"&File")
        self.menu_bar.Append(self.run_menu,"&Run")
        self.menu_bar.Append(self.help_menu,"&Help")

        #add the menu bar to the window
        self.SetMenuBar(self.menu_bar)

    def enable_close_tab(self):
        self.menu_close_tab.Enable(True)
        self.tool_bar.EnableTool(self.tb_close.GetId(),True)
    def disable_close_tab(self):
        self.menu_close_tab.Enable(False)
        self.tool_bar.EnableTool(self.tb_close.GetId(),False)
    def enable_run(self):
        self.menu_run.Enable(True)
        self.tool_bar.EnableTool(self.tb_run.GetId(),True)
    def disable_run(self):
        self.menu_run.Enable(False)
        self.tool_bar.EnableTool(self.tb_run.GetId(),False)
    def enable_step(self):
        self.menu_step.Enable(True)
        self.tool_bar.EnableTool(self.tb_step.GetId(),True)
    def disable_step(self):
        self.menu_step.Enable(False)
        self.tool_bar.EnableTool(self.tb_step.GetId(),False)
    def enable_abort(self):
        self.menu_abort.Enable(True)
        self.tool_bar.EnableTool(self.tb_abort.GetId(),True)
    def disable_abort(self):
        self.menu_abort.Enable(False)
        self.tool_bar.EnableTool(self.tb_abort.GetId(),False)
        
#I'll just assume that the menu event handlers are self-explanatory

    def pause(self):
        if self.is_running:
            self.main_pipe.send("P")
            self.enable_step()

    def resume(self):
        if self.is_running:
            self.main_pipe.send("R")
            self.disable_step()

    def OnCloseTab(self,event):
        self.close_current_tab()
        if HighlighterPanel.PANEL_COUNT < 1:
            self.disable_close_tab()
            self.disable_run()
    
    def run_current_tab(self):
        if self.is_running:
            d = wx.MessageDialog(self,message="Do you want to abort the currently monitored program and run your new selection?",caption="Abort and restart?",style=wx.YES_NO|wx.YES_DEFAULT|wx.ICON_QUESTION)
            if d.ShowModal() != wx.ID_YES:
                return
            self.main_pipe.send("D")
        
        #Create the monitor and the pipe to it
        mypipe,hispipe = processing.Pipe()
        self.main_pipe = mypipe
        self.monitor = HighlighterMonitor(hispipe)
        
        #A hack to get console stuff working
        out_conside,out_monside = processing.Pipe()
        self.monitor.set_stdout_pipe(out_monside)
        self.console.set_stdout_pipe(out_conside)
        in_conside,in_monside = processing.Pipe()
        self.monitor.set_stdin_pipe(in_monside)
        self.console.set_stdin_pipe(in_conside)
        
        cur_panel = self.notebook.GetCurrentPage()
        file_path = cur_panel.text_file_path
        
        self.register_pages_with_monitor()
        
        self.monitor.set_module_to_run(file_path)
        self.monitor.start()

        self.is_running = True

        self.OnTBSliderUpdate(None) #Just so the state of slowing gets sent at the start
        
    def OnRunCurTab(self,event):
        self.run_current_tab()
        #self.disable_run()
        self.enable_abort()

    def close_current_tab(self):
        HighlighterPanel.PANEL_COUNT -= 1
        cur_tab = self.notebook.GetCurrentPage()
        cur_tab.die()
        self.notebook.DeletePageByReference(cur_tab)
        
    def register_pages_with_monitor(self):
        for i in range(self.notebook.GetPageCount()):
            page = self.notebook.GetPage(i)
            self.send_page_to_monitor(page)

    def send_page_to_monitor(self, page):
        page_pipe, mon_pipe = processing.Pipe()
        page.set_pipe(page_pipe)
        self.main_pipe.send("N")
        self.main_pipe.send(page.text_file_path)
        self.main_pipe.send(mon_pipe)
    
    def OnOpen(self, event):
        open_dialogue = wx.FileDialog(self,message="Open Python source files for highlighting",
                defaultDir=os.getcwd(), defaultFile="", wildcard="Python source files (*.py;*.pyw)|*.py;*.pyw", style=wx.OPEN|wx.MULTIPLE|wx.CHANGE_DIR)
        if open_dialogue.ShowModal() == wx.ID_OK:
            paths = open_dialogue.GetPaths()
            self.open_files(paths)

    def open_files(self, paths):
        oldpc = HighlighterPanel.PANEL_COUNT
        for path in paths:
            if self.file_is_valid(path):
                file_name = path[path.rindex(os.sep)+1:]
                newpage = HighlighterPanel(self.notebook, path, self.notebook, self.img_count)
                self.notebook.AddPage(newpage,file_name)
                self.notebook.SetPageImageByReference(newpage,0)
                if self.is_running:
                    self.send_page_to_monitor(newpage)
                self.notebook.SetSelection(self.notebook.GetPageCount()-1)
        if oldpc < 1 and HighlighterPanel.PANEL_COUNT > 0:
            if not self.is_running:
                self.enable_run()
            self.enable_close_tab()

    def file_is_valid(self, filename):
        #make sure the file is .py, .pyw, or has no extension
        if not (filename.endswith(".py") or filename.endswith(".pyw")):
            wx.MessageDialog(None, caption="An error has occurred!", message="Whoops! It appears that %s isn't a valid Python module! Make sure it is a valid Python file and has a proper file extension.\n\nTo run a file in PyLighter, its extension must be .py or .pyw." % filename, style=wx.ICON_ERROR|wx.OK).ShowModal()
            return False
        #try reading the file to make sure nothing is wrong with it
        try:
            file = open(filename, 'r')
            text = file.read()
            test_ctrl = wx.TextCtrl(self,-1)
            test_ctrl.Hide()
            test_ctrl.WriteText(text)
            file.close()
        except IOError, error:
            wx.MessageDialog(None, caption="An error has occurred!", message="Whoops! It appears that I can't open \"%s\"!\n\nHere's Python's error message:\n\n%s" % (filename, str(error)), style=wx.ICON_ERROR|wx.OK).ShowModal()
            return False
        except UnicodeDecodeError, error:
            wx.MessageDialog(None, caption="An error has occurred!", message="Whoops! It appears that I can't decode the file \"%s\"! Are you sure it's a Python file?\n\nHere's Python's error message:\n\n%s" % (filename, str(error)), style=wx.ICON_ERROR|wx.OK).ShowModal()
            return False
        else:
            return True

    def OnClose(self, event):
        d = wx.MessageDialog(self,message="Do you want to abort the currently monitored program and quit?",caption="Abort and quit?",style=wx.YES_NO|wx.YES_DEFAULT|wx.ICON_QUESTION)
        if d.ShowModal() != wx.ID_YES:
            return
        if self.is_running:
            self.main_pipe.send("R")
            self.main_pipe.send("SD")
            self.main_pipe.send("D")
        sys.exit(0)

    def OnAbout(self,event):
        wx.AboutBox(self.about_info)

    def OnStep(self,event):
        if self.is_running:
            self.main_pipe.send("ST")

    def OnAbort(self,event):
        d = wx.MessageDialog(self,message="Do you want to abort the currently monitored program?",caption="Abort?",style=wx.YES_NO|wx.YES_DEFAULT|wx.ICON_QUESTION)
        if d.ShowModal() != wx.ID_YES:
            return
        
        if self.is_running:
            self.main_pipe.send("R")
            self.main_pipe.send("D")
        self.disable_abort()

    def set_slowing_duration(self, dur):
        self.slowing_duration = dur
        if self.is_running:
            self.main_pipe.send("SSD")
            self.main_pipe.send(self.slowing_duration)

    def disable_slowing(self):
        if self.is_running:
            self.main_pipe.send("SD")
        
    def enable_slowing(self):
        if self.is_running:
            self.main_pipe.send("SE")

class TabOpenDropTarget(wx.FileDropTarget):
    def __init__(self, boss):
        wx.FileDropTarget.__init__(self)
        self.boss = boss
    
    def OnDropFiles(self, x, y, filenames):
        self.boss.open_files(filenames)
